
object CPoint {
  
  def cartesian(x:Double,y:Double,c:Color):CPoint =
    new CartesianCPoint(x,y,c)
  
  def polar(r:Double,theta:Double,c:Color):CPoint = 
    if (r<0) 
      throw new Error("r negative") 
    else 
      new PolarCPoint(r,theta,c)


  // reconstructCart
  // reconstructPolar



  private class CartesianCPoint (xpos:Double, ypos:Double, c:Color) 
         extends CartesianPoint(xpos,ypos) with CPoint {


    def distance (q:CPoint):Double = super[CartesianPoint].distance(q)


    // these method cannot be inherited

    override def move (dx:Double,dy:Double):CPoint = 
      new CartesianCPoint(xpos+dx, ypos+dy,c)

    def add (q:CPoint):CPoint =
      new CartesianCPoint(xpos+q.xCoord(),ypos+q.yCoord(),q.color())

    override def rotate (t:Double):CPoint = 
      new CartesianCPoint(xpos*math.cos(t)-ypos*math.sin(t),
 			  xpos*math.sin(t)+ypos*math.cos(t),
			  c)
    
    def isEqual (q:CPoint):Boolean = 
      (xpos == q.xCoord()) && (ypos == q.yCoord()) && (c==q.color())


    // Specific to color points
    
    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new CartesianCPoint(xpos,ypos,nc)


    // BRIDGE METHODS

    override def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => isEqual(cq)
      case _ => false
    }

    // CANONICAL METHODS

    override def toString ():String = 
      "cartesian(" + xpos + "," + ypos + "," + c + ")"

    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
        41 * (
  	  41 + xpos.hashCode()
        ) + ypos.hashCode()
      ) + c.hashCode()
  }



  private class PolarCPoint (r:Double, theta:Double, c:Color) 
            extends PolarPoint(r,theta) with CPoint {


    def distance (q:CPoint):Double = super[PolarPoint].distance(q)


    // these method cannot be inherited

    override def move (dx:Double,dy:Double):CPoint = 
      new CartesianCPoint(xCoord()+dx, yCoord()+dy,c)
    
    def add (q:CPoint):CPoint = 
      new CartesianCPoint(xCoord()+q.xCoord(),yCoord()+q.yCoord(),q.color())

    override def rotate (angle:Double):CPoint = 
      new PolarCPoint(r, theta+angle,c)

    private def normalize (angle:Double):Double = 
      if (angle >= 2*math.Pi)
	normalize(angle-2*math.Pi)
      else if (angle < 0)
	normalize(angle+2*math.Pi)
      else
	angle

    def isEqual (q:CPoint):Boolean = {
      r==q.distanceFromOrigin() && 
      normalize(theta)==normalize(q.angleWithXAxis()) &&
      c==q.color()
    }


    // Specific to color points
   
    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new PolarCPoint(r,theta,nc)
      
    // BRIDGE METHODS

    override def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => isEqual(cq)
      case _ => false
    }

    // CANONICAL METHODS
    
    override def toString ():String = 
      "polar(" + r + "," + theta + "," + c + ")"
    
    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
        41 * (
  	  41 + r.hashCode()
        ) + theta.hashCode()
      ) + c.hashCode()
  }
}



trait CPoint extends Point {

  def xCoord ():Double
  def yCoord ():Double
  def angleWithXAxis ():Double
  def distanceFromOrigin ():Double
  def distance (q:CPoint):Double
  def move (dx:Double,dy:Double):CPoint
  def add (q:CPoint):CPoint
  def rotate (theta:Double):CPoint
  def isEqual (q:CPoint):Boolean
  def isOrigin ():Boolean

  def color ():Color
  def updateColor (nc:Color):CPoint

  // bridge methods
  def distance (q:Point):Double
  def add (q:Point):Point
  def isEqual (q:Point):Boolean
}
