
object CPoint {
  
  def cartesian(x:Double,y:Double,c:Color):CPoint =
    new CartesianCPoint(x,y,c)
  
  def polar(r:Double,theta:Double,c:Color):CPoint = 
    if (r<0) 
      throw new Error("r negative") 
    else 
      new PolarCPoint(r,theta,c)


  // helper functions to simplify some inheritance

  private def reconstructCart (p:Point,c:Color):CPoint =
      new CartesianCPoint(p.xCoord(),p.yCoord(),c)

  private def reconstructPolar (p:Point,c:Color):CPoint =
      new PolarCPoint(p.distanceFromOrigin(),
                      p.angleWithXAxis(),c)



  private class CartesianCPoint (xpos:Double, ypos:Double, c:Color) 
             extends CartesianPoint(xpos,ypos) with CPoint {

    def distance (q:CPoint):Double = 
      super[CartesianPoint].distance(q)

    override def move (dx:Double,dy:Double):CPoint = 
      reconstructCart(super[CartesianPoint].move(dx,dy),c)

    def add (q:CPoint):CPoint = 
      reconstructCart(super[CartesianPoint].add(q),q.color())

    override def rotate (t:Double):CPoint = 
      reconstructCart(super[CartesianPoint].rotate(t),c)

    def isEqual (q:CPoint):Boolean = 
      super[CartesianPoint].isEqual(q) && (c==q.color())
    
    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new CartesianCPoint(xpos,ypos,nc)


    // BRIDGE METHODS

    override def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => isEqual(cq)
      case _ => false
    }


    // CANONICAL METHODS

    override def toString ():String = 
      "cartesian(" + xpos + "," + ypos + "," + c + ")"

    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
        41 * (
  	  41 + xpos.hashCode()
        ) + ypos.hashCode()
      ) + c.hashCode()
  }



  private class PolarCPoint (r:Double, theta:Double, c:Color) 
            extends PolarPoint(r,theta) with CPoint {

    def distance (q:CPoint):Double = 
      super[PolarPoint].distance(q)

    override def move (dx:Double,dy:Double):CPoint = 
      reconstructCart(super[PolarPoint].move(dx,dy),c)
    
    def add (q:CPoint):CPoint = 
      reconstructCart(super[PolarPoint].add(q),q.color())

    override def rotate (angle:Double):CPoint = 
      reconstructPolar(super[PolarPoint].rotate(angle),c)

    def isEqual (q:CPoint):Boolean = {
      super[PolarPoint].isEqual(q) && c==q.color()
    }

    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new PolarCPoint(r,theta,nc)
      
    // BRIDGE METHODS

    override def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => isEqual(cq)
      case _ => false
    }

    // CANONICAL METHODS
    
    override def toString ():String = 
      "polar(" + r + "," + theta + "," + c + ")"
    
    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
        41 * (
  	  41 + r.hashCode()
        ) + theta.hashCode()
      ) + c.hashCode()
  }
}



trait CPoint extends Point {

  def xCoord ():Double
  def yCoord ():Double
  def angleWithXAxis ():Double
  def distanceFromOrigin ():Double
  def distance (q:CPoint):Double
  def move (dx:Double,dy:Double):CPoint
  def add (q:CPoint):CPoint 
  def rotate (theta:Double):CPoint
  def isEqual (q:CPoint):Boolean 
  def isOrigin ():Boolean

  def color ():Color
  def updateColor (nc:Color):CPoint

  // bridge methods
  def distance (q:Point):Double
  def add (q:Point):Point
  def isEqual (q:Point):Boolean
}
